/***************************************************************
 *
 * NAME:
 * 	masterplan.c
 *
 * DESCRIPTION:
 *	A who-is-fingering-me program.
 *
 * COPYRIGHT:
 *	Copyright (C) 1994  Laurion Burchall <ldb@netspace.org>
 *
 * AUTHOR:
 * 	Laurion Burchall
 *
 * EMAIL:
 * 	ldb@netspace.org
 *
 * HISTORY:
 *      August, 1994	Laurion Burchall <ldb@netspace.org>
 *                      created
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 ***************************************************************/

/*****************************************************************
 * I N C L U D E S
 ****************************************************************/
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>

#include <signal.h>
#include <errno.h>
#include <fcntl.h>
#include <pwd.h>

#include "masterplan.h"
#include "config.h"

#ifndef PATH_MAX
#ifndef _POSIX_PATH_MAX
#define PATH_MAX 1024
#else
#define PATH_MAX _POSIX_PATH_MAX
#endif
#endif

/*****************************************************************
 * P R O T O T Y P E S
 ****************************************************************/
static void justOne (const char * pidFile);
static void stillRunning (const char * pidFile);

static int startup(int argc, char *argv[]);

static int createPipePlan (void);
static int checkSymPlan (void);

static void writePlan (int fd);

static void writeVar (int fd, FILE * plan);
static void writeProg (int fd, FILE * plan);

/*
 * these are the functions that return the variable strings
 */
static char * getNumFingers(void);
static char * getDay(void);
static char * getMonths(void);
static char * getMonthStr(void);
static char * getYear(void);
static char * getDate(void);
static char * getTime(void);

/*
 * these are the signal handler functions
 */
static void sigDie(int sig);
static void sigCld(int sig);

/*****************************************************************
 * T Y P E S
 ****************************************************************/
/*
 * This table maps a variable word to the function that
 * returns its string. The array should be terminated with
 * a null entry.
 */
typedef char*(*VARFUNC)(void);
typedef struct varFunc {
    const char * name;
    VARFUNC func;
} varFunc;

/*****************************************************************
 * G L O B A L S (gasp!)
 ****************************************************************/
/*
 * the name of the program
 */
static char * pname = 0;

/*
 * Are we in verbose mode
 */
static int Verbose = 0;

/*
 * the number of times we have been fingered. the implementation
 * of the counting scheme is open to race conditions but the worst
 * damage they can do is cause the finger count to be too low.
 */
static int  NumFingers = 0;

/* 
 * The files that we use
 * add some extra space for safety
 */
static char PlanFile[PATH_MAX+5];
static char PipeFile[PATH_MAX+5];
static char PidFile[PATH_MAX+5];
static char PlanTemplateFile[PATH_MAX+5];

/* The variables that we know */
static varFunc VarFuncs[] = {  
    { "FINGERS", getNumFingers },
    { "DAY", 	getDay },
    { "MONTHS", getMonths },
    { "MONTH", 	getMonthStr },
    { "YEAR", 	getYear },
    { "DATE", 	getDate },
    { "TIME",   getTime },
    { 0, 0 }
};

/****************************************************************
 * FUNCTION: justOne
 *
 * This makes sure that only one copy of the given program is 
 * running. It uses the pidFile passed to it to do that.
 *
 ***************************************************************/
static void justOne ( const char * pidFile )
{
    int fd = -1;
    char   buf[1024];
    
    if ((fd = open(pidFile, O_RDONLY)) < 0)
	goto notrunning;

    if (read(fd, buf, 1024) < 0) {
	    goto notrunning;
    }

    if (kill((pid_t)atoi(buf), 0) < 0)
	{
	    switch (errno)
		{
		case EPERM:
		    goto notrunning;
		    /* NOTREACHED */
		    break;
		case ESRCH:
		    goto notrunning;
		    /* NOTREACHED */
		    break;
		default:
		    break;
		}
	}

    (void)fprintf(stderr, "%s: already running. pid %d\n", pname, atoi(buf));
    exit(0);

  notrunning:
    if ((fd >= 0) && (close(fd) < 0)) {
	(void)fprintf(stderr, "%s: close: %s\n", pname, strerror(errno));
    }
    if ((fd = creat(pidFile, 0600)) < 0) {
	(void)fprintf(stderr, "%s: creat(%s): %s\n", pname, pidFile, 
		strerror(errno));
	exit(1);
    }
    (void)sprintf(buf, "%d", (int)getpid());
    if (write(fd, buf, strlen(buf)) < 0) {
	(void)fprintf(stderr, "%s: write: %s\n", pname, strerror(errno));
	exit(1);
    }
    if (close(fd) < 0) {
	(void)fprintf(stderr, "%s: close: %s\n", pname, strerror(errno));
	exit(1);
    }
}


/****************************************************************
 * FUNCTION: stillRunning
 *
 * This makes sure that the pid file still exists and shows that
 * the program is still running.
 *
 ***************************************************************/
static void stillRunning ( const char * pidFile )
{
    int fd = -1;
    char   buf[1024];
    
    if ((fd = open(pidFile, O_RDONLY)) < 0)
	goto notrunning;

    if (read(fd, buf, 1024) < 0) {
	goto notrunning;
    }

    if ((pid_t)atoi(buf) != getpid()) {
	goto notrunning;
    }

    if (close(fd) < 0) {
	(void)fprintf(stderr, "%s: close: %s\n", pname, strerror(errno));
    }

    return;

  notrunning:
    if ((fd >= 0) && (close(fd) < 0)) {
	(void)fprintf(stderr, "%s: close: %s\n", pname, strerror(errno));
    }
    (void)fprintf(stderr, "%s: PID file changed. exiting.\n", pname);
    exit(0);
}

/****************************************************************
 * FUNCTION: checkSymPlan
 *
 * make sure that the plan exists and is the right symbolic link
 *
 ***************************************************************/
static int checkSymPlan (void)
{
    char buf[PATH_MAX];

    /* read in the symlink */
    if (readlink(PlanFile, buf, PATH_MAX) < 0) {
	switch (errno) {
	case EINVAL:
	    (void)fprintf(stderr, "%s: %s is not a symbolic link!\n",
		    pname, PlanFile);
	    return 1;
	    /* NOTREACHED */
	    break;
	default:
	    break;
	}
    } else {
	/* the link exists. see if it points where we want */

	if (!strncmp(buf, PipeFile, strlen(PipeFile)))
	    return 0;
	if (remove(PlanFile) < 0) {
	    (void)fprintf(stderr, "%s: remove(%s): %s\n", pname,
		    PlanFile, strerror(errno));
	    return 1;
	}
    }

    /* create the symlink */
    if (symlink(PipeFile, PlanFile) < 0) {
	(void)fprintf(stderr, "%s: symlink(%s, %s): %s\n", pname, 
		PipeFile, PlanFile, strerror(errno));
	return 1;
    }
    return 0;
}

/****************************************************************
 * FUNCTION: createPipPlan
 *
 * Makes the pipe whose name is in PipeFile
 *
 ***************************************************************/
static int createPipePlan (void)
{
    /*
     * we remove the pipe file incase it exists
     * we are the only copy running at this point (justOne()) so
     * it is O.K. to do this.
     */
    (void)remove(PipeFile);

    if (mknod(PipeFile, S_IFIFO | PERMS, 0)) {
	(void)fprintf(stderr, "%s: mknod(%s): %s\n", pname, 
		PipeFile, strerror(errno));
	return 1;
    }

    /* this seems to be necessary too */
    if (chmod(PipeFile, PERMS) < 0) {
	(void)fprintf(stderr, "%s: chmod(%s): %s\n", pname, 
		      PipeFile, strerror(errno));
	return 1;
    }
    return 0;
}

/****************************************************************
 * FUNCTION: writePlan
 *
 * writes the plan template into the pipe
 *
 ***************************************************************/
static void writePlan (int fd) {
    FILE * plan = 0;
    int i = 0;
    char c = 0;

    if ((plan = fopen(PlanTemplateFile, "r")) == 0) {
	switch(errno) {
	case ENOENT:
	    return;
	    /* NOTREACHED */
	    break;
	default:
	    break;
	}
	(void)fprintf(stderr, "%s: fopen(%s): %s", pname,
		PlanTemplateFile, strerror(errno));
    }
    while ((i = getc(plan)) != EOF) {
	c = i;
	switch(c) {
	case '$':
	    writeVar(fd, plan);
	    break;
	case '`':
	    writeProg(fd, plan);
	    break;
	case '\\':	/* this must come immediately before default */
	    if ((c = getc(plan)) == EOF) {
		goto done;
	    }
	    /* FALLTHRU */
	default:
	    if (write(fd, &c, 1) != 1) {
		(void)fprintf(stderr, "%s: write: %s\n", pname, 
			      strerror(errno));
		goto done;
	    }
	    break;
	}
    }
  done:
    if (fclose(plan) < 0) {
	(void)fprintf(stderr, "%s: fclose(%s): %s", pname,
		      PlanTemplateFile, strerror(errno));
    }
    return;
}

/****************************************************************
 * FUNCTION: writeVar
 *
 * obtains the variable and writes it into the pipe
 *
 ***************************************************************/
static void writeVar (int fd, FILE * plan) {
    char var[128];
    char c = 0;
    int i = 0;
    int n = 0;
    varFunc * v = VarFuncs;
    char * s = 0;
    
    /* we have read the inital $ so read until a whitespace */
    while (((i = getc(plan)) != EOF) && (isalnum(c = i))) {
	var[n++] = c;
    }
    if (ungetc(c, plan) == EOF) {
	(void)fprintf(stderr, "%s: ungetc: %s", pname, strerror(errno));
    }
    var[n] = 0;
    
    /* look for an internal variable */
    while (v->name) {
	if (!strcmp(var, v->name)) {
	    /* this is it */
	    s = v->func();
	    goto gotit;
	}
	v++;
    }

    /* we didn't find it */
    if (!(s = getenv(var))) {
	return;
    }
  gotit:
    n = strlen(s);
    if (write(fd, s, (unsigned)n) != n) {
	(void)fprintf(stderr, "%s: write: %s\n", pname,
		strerror(errno));
    }
    return;
}

/****************************************************************
 * FUNCTION: writeProg
 *
 * runs the program and puts its output into the pipe
 *
 ***************************************************************/
static void writeProg (int fd, FILE * plan) {
    char prog[PATH_MAX];
    int n = 0;
    int i = 0;
    char c = 0;
    pid_t proc;
    int stat;

    /* read until the next ` */
    while (((i = getc(plan)) != EOF) && ((c = i) != '`')) {
	prog[n++] = c;
    }
    prog[n] = 0;

    if ((proc = fork()) < 0) {
	(void)fprintf(stderr, "%s: fork: %s\n", pname, strerror(errno));
	return;
    }	
    if (proc == 0) {
	/* child */

	/*
	 * now we hook our stdout to the pipe
	 */
	if (close(STDOUT_FILENO) < 0) {
	    (void)fprintf(stderr, "%s (child): close: %s\n", 
			  pname, strerror(errno));
	    _exit(1);
	}
	if (dup2(fd, STDOUT_FILENO) < 0) {
	    (void)fprintf(stderr, "%s (child): dup2: %s\n", 
			  pname, strerror(errno));
	    _exit(1);
	}
	
	/* run the program */
	if (system(prog) < 0) {
	    (void)fprintf(stderr, "%s (child): system(%s): %s\n", pname,
		    prog, strerror(errno));
	}

	/* we must use _exit or the plan is closed too... */
	_exit(0);

    } else {
	/* parent */
	if (waitpid(proc, &stat, 0) < 0) {
	    if (errno != EINTR) {
		(void)fprintf(stderr, "%s (child): waitpid: %s\n", pname, 
			      strerror(errno));
	    }
	}
    }
    return;
}

/****************************************************************
 * FUNCTION: startup
 *
 * This does all the startup stuff. It returns 0 if the startup
 * was sucessful, 1 otherwise.
 *
 ***************************************************************/
static int startup(/*ARGSUSED*/ int argc, char *argv[])
{
    NumFingers = 0;
    justOne(PidFile);
    return 0;
}

/****************************************************************
 * FUNCTION: getNumFingers
 *
 * This returns a pointer to the string giving the current number
 * of times the user has been fingered
 *
 ****************************************************************/
static char * getNumFingers(void)
{
    static char buf[24];
    (void)sprintf(buf, "%d", NumFingers);
    return buf;
}

/****************************************************************
 * FUNCTION: getDay
 *
 * This returns a pointer to the string giving the current day of
 * the week.
 *
 ****************************************************************/
static char * getDay(void)
{
    static char buf[8];
    time_t clock = time(0);
    (void)strftime(buf, 8, "%a", localtime(&clock));
    return buf;
}

/****************************************************************
 * FUNCTION: getMonths
 *
 * This returns a pointer to the string giving the current month
 * number.
 *
 ****************************************************************/
static char * getMonths(void)
{
    static char buf[24];
    time_t clock = time(0);
    (void)strftime(buf, 24, "%m", localtime(&clock));
    return buf;
}

/****************************************************************
 * FUNCTION: getMonthStr
 *
 * This returns a pointer to the string giving the name of
 * the current month.
 *
 ****************************************************************/
static char * getMonthStr(void)
{
    static char buf[24];
    time_t clock = time(0);
    (void)strftime(buf, 24, "%b", localtime(&clock));
    return buf;
}

/****************************************************************
 * FUNCTION: getYear
 *
 * This returns a pointer to the string giving the current year
 *
 ****************************************************************/
static char * getYear(void)
{
    static char buf[8];
    time_t clock = time(0);
    (void)strftime(buf, 8, "%y", localtime(&clock));
    return buf;
}

/****************************************************************
 * FUNCTION: getDate
 *
 * This returns a pointer to the string giving the current date
 *
 ****************************************************************/
static char * getDate(void)
{
    static char buf[4];
    time_t clock = time(0);
    (void)strftime(buf, 4, "%e", localtime(&clock));
    return buf;
}

/****************************************************************
 * FUNCTION: getTime
 *
 * This returns a pointer to the string giving the current time
 *
 ****************************************************************/
static char * getTime(void)
{
    static char buf[10];
    time_t clock = time(0);
    /* The 12-hour clock people may want to use %I */
    (void)strftime(buf, 10, "%H:%M", localtime(&clock));
    return buf;
}

/****************************************************************
 * FUNCTION: sigDie
 *
 * This is called when a deadly signal is received.
 *
 ****************************************************************/
static void sigDie(int sig)
{
    (void)printf("%s: killed by signal %d\n", pname, sig);
    if (remove(PipeFile) < 0) {
	(void)fprintf(stderr, "%s: remove(%s): %s\n",
		pname, PipeFile, strerror(errno));
    }
    if (remove(PidFile) < 0) {
	(void)fprintf(stderr, "%s: remove(%s): %s\n",
		pname, PidFile, strerror(errno));
    }
    exit(0);
}

/****************************************************************
 * FUNCTION: sigCld
 *
 * This is called when a SIGCLD is received, it reaps the dead 
 * child.
 *
 ****************************************************************/
static void sigCld(/*ARGSUSED*/ int sig)
{
    
}

/****************************************************************
 * MAIN
 *
 ****************************************************************/
int main (int argc, char * argv[])
{
    struct passwd * pwent = 0;
    int c = 0;
    char * s = 0;
    int daemon = 0;

    pname = argv[0];

    if ((pwent = getpwuid(getuid())) == 0) {
	(void)fprintf(stderr, "%s: getpwuid: %s", pname, strerror(errno));
	exit(1);
    }

    (void)sprintf(PlanFile, "%s/%s", pwent->pw_dir, DEFAULT_PLANFILE);
    (void)sprintf(PipeFile, DEFAULT_PIPEFILE, pwent->pw_name);
    (void)sprintf(PidFile,  DEFAULT_PIDFILE, pwent->pw_name);
    (void)sprintf(PlanTemplateFile, "%s/%s", 
		  pwent->pw_dir, DEFAULT_MASTERPLANFILE);

    while ((c = getopt(argc, argv, "hdf:t:v")) != EOF) {
	switch(c) {
	case 'h':
	    /* FALLTHRU */
	case '?':
	    (void)fprintf(stderr, 
"usage: %s [OPTIONS]\n\n", pname);
	    (void)fprintf(stderr,
"   -d 	                run in the background\n");
	    (void)fprintf(stderr,
"   -f <file>           the file to write to\n");
	    (void)fprintf(stderr,
"   -t <template file>  the template file to use\n");
	    (void)fprintf(stderr,
"   -v                  verbose mode\n");
	    exit(0);
	    /* NOTREACHED */
	    break;
	case 'd':
	    daemon = 1;
	    break;
	case 'f':
	    (void)strcpy(PlanFile, optarg);
	    break;
	case 't':
	    (void)strcpy(PlanTemplateFile, optarg);
	    break;
	case 'v':
	    Verbose = 1;
	    break;
	default:
	    (void)fprintf(stderr, "%s: unknown option %c\n", pname, c);
	    exit(1);
	    /* NOTREACHED */
	    break;
	}
    }

    /* 
     * now append the name of our file to the extra files
     * we use. we will need this to run more than one copy
     * of masterplan
     */
    if ((s = strrchr(PlanFile, '/')) == 0) {
	s = PlanFile;
    } else {
	s++;
    }
    (void)strncat(PipeFile, s, PATH_MAX-strlen(PipeFile));
    (void)strncat(PidFile, s, PATH_MAX-strlen(PidFile));

    /*
     * print out the files we are going to use
     */
    if (Verbose) {
	(void)printf("PLANFILE: %s\n", PlanFile);
	(void)printf("PIPEFILE: %s\n", PipeFile);
	(void)printf("PIDFILE:  %s\n", PidFile);
	(void)printf("TEMPLATE: %s\n", PlanTemplateFile);
    }

    /*
     * become a daemon if necessary
     */
    if (daemon) {
	pid_t proc;
	if ((proc = fork()) < 0) {
	    (void)fprintf(stderr, "%s: fork: %s",
			  pname, strerror(errno));
	    exit(1);
	}
	if (proc != 0) {
	    exit(0);
	}
	if (signal(SIGHUP, SIG_IGN) == SIG_ERR) {
	    (void)fprintf(stderr, "%s: signal: %s\n", pname, strerror(errno));
	}
    }

    /* startup */
    if (startup(argc, argv) || createPipePlan() || checkSymPlan()) {
	exit(1);
    }

    /*
     * Signal handling
     *
     * we will only exit if killed so we must arrange to clean up
     */

    /* Catch signals from dead children */
    if (signal(SIGCHLD, sigCld) == SIG_ERR) {
	(void)fprintf(stderr, "%s: signal: %s\n", pname, strerror(errno));
    }

    /* catch signals that will kill us so we can clean up */
    if (signal(SIGINT, sigDie) == SIG_ERR) {
	(void)fprintf(stderr, "%s: signal: %s\n", pname, strerror(errno));
    }
    if (signal(SIGHUP, sigDie) == SIG_ERR) {
	(void)fprintf(stderr, "%s: signal: %s\n", pname, strerror(errno));
    }
    if (signal(SIGQUIT, sigDie) == SIG_ERR) {
	(void)fprintf(stderr, "%s: signal: %s\n", pname, strerror(errno));
    }
    if (signal(SIGTERM ,sigDie) == SIG_ERR) {
	(void)fprintf(stderr, "%s: signal: %s\n", pname, strerror(errno));
    }

    /* these are so we clean up even if we crash */
    if (signal(SIGSEGV, sigDie) == SIG_ERR) {
	(void)fprintf(stderr, "%s: signal: %s\n", pname, strerror(errno));
    }
    if (signal(SIGBUS, sigDie) == SIG_ERR) {
	(void)fprintf(stderr, "%s: signal: %s\n", pname, strerror(errno));
    }

    /* Ignore broken pipe errors - just continue */
    if (signal(SIGPIPE, SIG_IGN) == SIG_ERR) {
	(void)fprintf(stderr, "%s: signal: %s\n", pname, strerror(errno));
    }

    /*
     * the main loop
     *
     * this opens the pipe and reads. when the read succeeds the 
     * plan template is written into it
     */

    while (/*CONSTCOND*/ 1) {
	int fd = -1;
	stillRunning(PidFile);
	if ((fd = open(PipeFile, O_WRONLY)) < 0) {
	    (void)fprintf(stderr, "%s: open(%s): %s\n", pname,
		    PipeFile, strerror(errno));
	    exit(1);
	}
	if (write(fd, 0, 0) < 0) {
	    (void)fprintf(stderr, "%s: write: %s\n", pname, strerror(errno));
	    exit(1);
	}
	if (Verbose) {
	    (void)printf("%s: pipe being read\n", pname);
	}
	NumFingers++;
	writePlan(fd);
	if (close(fd) < 0) {
	    (void)fprintf(stderr, "%s: close: %s\n", pname, strerror(errno));
	}

	/* 
	 * we sleep to avoid filling the same request twice 
	 * and to slow down responses to constant fingering
	 */
	(void)sleep(1);
    }
    /*NOTREACHED*/
    return 0;
}
